from __future__ import absolute_import, unicode_literals

import six
import pytest
from pytest import approx

from barril._foundation.odict import odict
from barril import units
from barril.units import InvalidUnitError, ObtainQuantity, Quantity


def testFormatting(unit_database_start_units):
    point = units.FixedArray(2, "length", [(100, 150), (50, 50)], "m")
    assert "(100, 150) (50, 50) [m]" == six.text_type(point)


def testValues(unit_database_start_units):
    with pytest.raises(ValueError):
        units.FixedArray(1, "length", [100], "m")
    point = units.FixedArray(3, "length", [100, 150, 200], "m")
    assert point.unit == "m"
    assert point.values == [100, 150, 200]
    assert point.dimension == 3

    # check dimension read-only
    def SetDimensions():
        point.dimension = 4

    with pytest.raises(AttributeError):
        SetDimensions()

    # check number of values
    with pytest.raises(ValueError):
        point.CreateCopy(values=[0, 0])
    with pytest.raises(ValueError):
        point.CreateCopy(values=[0, 0, 0, 0])
    point = point.CreateCopy(values=[50, 80, 90])  # OK

    # check conversion
    assert point.GetValues("m") == [50, 80, 90]

    assert approx(point.GetValues("km")) == [0.050, 0.080, 0.090]

    assert "50 80 90 [m]" == six.text_type(point)

    with pytest.raises(AttributeError):
        setattr(point, "values", [50, 80, 90])


def testInvalidUnits(unit_database_start_units):
    point = units.FixedArray(2, "length", [(100, 150), (50, 50)], "m")
    try:
        point = point.CreateCopy(values=[(20, 30), (40, 50)], unit="foo")
    except InvalidUnitError:
        pass

    assert point.values == [(100, 150), (50, 50)]
    assert point.unit == "m"


def testCreateWithQuantity(unit_database_start_units):
    units.FixedArray.CreateWithQuantity(
        Quantity.CreateDerived(odict()), [100, 150, 200], dimension=3
    )

    quantity = ObtainQuantity("m", "length")
    a1 = units.FixedArray(3, quantity, values=[1, 2, 3])
    a2 = units.FixedArray(3, quantity, values=[1, 2, 3])

    assert a1.GetValues("km") == [0.001, 0.002, 0.003]
    assert a2.GetValues() == [1, 2, 3]


def testCopy(unit_database_start_units):
    point = units.FixedArray(3, "length", [100, 150, 200], "m")
    assert point.Copy() == point

    #         other = units.FixedArray(3, 'length', [1, 2, 3], 'km')
    other = point.CreateCopy()
    assert other == point

    # test copy with subclasses
    class MyFixedArray(units.FixedArray):
        pass

    point = MyFixedArray(3, "length", [100, 150, 200], "m")
    assert isinstance(point.Copy(), MyFixedArray)

    point = units.FixedArray(3, "length", [100, 150, 200], "m")


def testEmptyArray(unit_database_start_units):
    arr = units.FixedArray.CreateEmptyArray(3)
    assert not arr.HasCategory()
    assert arr.category == ""
    assert arr.unit == ""

    arr = arr.CreateCopy(category="length", unit="m")
    assert arr.HasCategory()
    assert arr.category == "length"
    assert arr.unit == "m"


def testReadOnlyQuantity(unit_database_start_units):
    quantity = ObtainQuantity("m", "length")
    array = units.FixedArray(3, quantity, values=[1, 2, 3])
    assert approx(array.GetValues("km")) == [0.001, 0.002, 0.003]


def testDefaultValues(unit_database_start_units):
    # Not raises exception because by default validation is False on the copy operation
    array = units.FixedArray(3, "length")
    assert array.values == [0.0, 0.0, 0.0]

    array = units.FixedArray(3, ObtainQuantity("m"))
    assert array.values == [0.0, 0.0, 0.0]

    with pytest.raises(AssertionError):
        units.FixedArray(3, ObtainQuantity("m"), unit="m")


def testTupleCreation(unit_database_start_units):
    """
    Make sure we can use tuples as values.
    """
    array = units.FixedArray(3, (0.0, 200.0, 300.0), "m")
    assert units.FixedArray(3, "length", (0.0, 200.0, 300.0), "m") == array


def testFixedArrayPickle(unit_database_start_units):
    import pickle

    fixed_array_1 = units.FixedArray(3, "length", values=[1, 2, 3], unit="m")
    fixed_array_2 = pickle.loads(pickle.dumps(fixed_array_1))

    assert fixed_array_1 == fixed_array_2


def testNumberInteractions():
    # Operations on FixedArrays used to return Array instances.
    a = units.FixedArray(3, [1, 2, 3], "m")
    b = units.FixedArray(3, [0.5] * 3, "m")

    s = a + b
    assert isinstance(s, units.FixedArray)
    assert s.GetDimension() == 3

    m = a + b
    assert isinstance(m, units.FixedArray)
    assert m.GetDimension() == 3


def testFixedArrayChangingIndex():
    fixed_array = units.FixedArray(3, [1, 2, 3], "m")
    assert fixed_array.ChangingIndex(0, 5) == units.FixedArray(3, [5, 2, 3], "m")
    assert fixed_array.ChangingIndex(0, units.Scalar(5, "m")) == units.FixedArray(
        3, [5, 2, 3], "m"
    )
    assert fixed_array.ChangingIndex(0, (5,)) == units.FixedArray(3, [5, 2, 3], "m")

    # Different unit
    assert fixed_array.ChangingIndex(0, (5, "cm")) == units.FixedArray(
        3, [5, 200, 300], "cm"
    )
    assert fixed_array.ChangingIndex(0, units.Scalar(5, "cm")) == units.FixedArray(
        3, [5, 200, 300], "cm"
    )

    # Keeping fixed array unit
    assert fixed_array.ChangingIndex(
        0, units.Scalar(5, "cm"), use_value_unit=False
    ) == units.FixedArray(3, [0.05, 2, 3], "m")

    # Other indexes
    assert fixed_array.ChangingIndex(1, units.Scalar(5, "cm")) == units.FixedArray(
        3, [100, 5, 300], "cm"
    )
    assert fixed_array.ChangingIndex(2, units.Scalar(5, "cm")) == units.FixedArray(
        3, [100, 200, 5], "cm"
    )


def testFixedArrayIndexAsScalar():
    from barril.units import Scalar

    fixed_array = units.FixedArray(3, "length of path", [1, 2, 3], "m")
    assert fixed_array.IndexAsScalar(0).GetCategory() == "length of path"
    assert fixed_array.IndexAsScalar(0) == Scalar("length of path", 1, "m")
    assert fixed_array.IndexAsScalar(1) == Scalar("length of path", 2, "m")



from __future__ import unicode_literals

# Basic extensions to the python language or standard library.
# As a rule of thumb, objects and functions here shouldn't depend
# on any other modules from package.



from __future__ import unicode_literals

from barril._foundation.klass import AllBasesNames, IsInstance, IsSubclass


class _A(object):
    pass


class _B(object):
    pass


class _C(_B):
    pass


class _D(_C, _A):
    pass


class _E(_D, _C, _A):
    pass


def testIsInstance():
    assert IsInstance(_C(), "_B")
    assert IsInstance(_C(), ("_B",))
    assert not IsInstance(_C(), ("_A",))
    assert IsInstance(_C(), ("_A", "_B"))
    assert not IsInstance(_C(), ("_A", "_D"))


def testIsSubclass():
    assert IsSubclass(_C, ("_C",))
    assert IsSubclass(_C, (_C.__name__,))
    assert IsSubclass(_C, "_B")
    assert IsSubclass(_C, _B.__name__)
    assert IsSubclass(_C, ("_B",))
    assert IsSubclass(_C, (_B.__name__,))
    assert not IsSubclass(_C, ("_A",))
    assert not IsSubclass(_C, (_A.__name__,))
    assert IsSubclass(_C, ("_A", "_B"))
    assert not IsSubclass(_C, ("_A", "_D"))
    assert not IsSubclass(_A, ("_C",))


def testKlass():
    class A(object):
        pass

    class B(A):
        pass

    class C(B):
        pass

    class D(C):
        pass

    class Alpha(object):
        pass

    class AlphaC(Alpha, C):
        pass

    assert set(AllBasesNames(A)) == {"object"}
    assert set(AllBasesNames(B)) == {"A", "object"}
    assert set(AllBasesNames(C)) == {"B", "A", "object"}
    assert set(AllBasesNames(D)) == {"C", "B", "A", "object"}
    assert set(AllBasesNames(AlphaC)) == {"Alpha", "object", "C", "B", "A"}

    assert IsInstance(A(), "object")
    assert IsInstance(A(), "A")
    assert IsInstance(B(), "object")
    assert IsInstance(B(), "A")
    assert IsInstance(B(), "B")
    assert IsInstance(C(), "object")
    assert IsInstance(C(), "A")
    assert IsInstance(C(), "B")
    assert IsInstance(C(), "C")
    assert IsInstance(AlphaC(), "object")
    assert IsInstance(AlphaC(), "A")
    assert IsInstance(AlphaC(), A)
    assert IsInstance(AlphaC(), "B")
    assert IsInstance(AlphaC(), "C")
    assert IsInstance(AlphaC(), "Alpha")
    assert IsInstance(AlphaC(), "AlphaC")
    assert IsInstance(AlphaC(), AlphaC)

    assert not IsInstance(AlphaC(), "Rubles")
    assert not IsInstance(A(), "B")
    assert not IsInstance(B(), "C")



# coding=utf-8
from __future__ import unicode_literals

import traceback

import pytest
import six

from barril._foundation.exceptions import ExceptionToUnicode
from barril._foundation.reraise import Reraise

pytestmark = pytest.mark.skipif(
    condition=not six.PY2,
    reason="On Python 3, Reraise is just a redirection to Python's raise...from statement",
)


# ===================================================================================================
# ExecutePythonCode
# ===================================================================================================
def ExecutePythonCode(code):
    exec(compile(code, "<string>", "exec"))


# ===================================================================================================
# ExceptionTestConfiguration
# ===================================================================================================
class ExceptionTestConfiguration:
    def __init__(
        self,
        exception_type,
        string_statement,
        expected_inner_exception_message="",
        expected_traceback_message=None,
    ):
        self.exception_type = exception_type
        self.string_statement = string_statement
        self.expected_inner_exception_message = expected_inner_exception_message
        self.expected_traceback_message = expected_traceback_message

    def RaiseExceptionUsingReraise(self):
        def RaiseException():
            ExecutePythonCode(self.string_statement)
            pytest.fail("Should not reach here")

        def ReraiseException():
            try:
                RaiseException()
            except self.exception_type as exception:
                Reraise(exception, "While doing 'bar'")

        try:
            try:
                ReraiseException()
            except self.exception_type as e1:
                Reraise(e1, "While doing x:")
        except self.exception_type as e2:
            Reraise(e2, "While doing y:")

    def GetExpectedExceptionMessage(self):
        return (
            "\nWhile doing y:\nWhile doing x:\nWhile doing 'bar'\n"
            + self.expected_inner_exception_message
        )

    def GetExpectedTracebackMessage(self, actual_exception, use_utf_8_encoding=True):
        # Getting the type of the "actual exception" because its type might be different than self.exception_type
        reraised_exception_name = type(actual_exception).__name__
        exception_message = self.GetExpectedExceptionMessage()

        if self.expected_traceback_message is not None:
            exception_message = self.expected_traceback_message
        else:
            exception_message = (
                reraised_exception_name + ": " + exception_message + "\n"
            )

        # HACK [muenz]: the 'traceback' module does this, so in order to be able to compare strings we
        # need this workaround notice that if the exception "leaks" the Python console handles the
        # unicode symbols properly
        if use_utf_8_encoding:
            exception_message = exception_message.encode("utf-8")
        else:
            exception_message = traceback._some_str(exception_message)
        assert isinstance(exception_message, bytes)

        return exception_message


parametrized_exceptions = pytest.mark.parametrize(
    "exception_configuration",
    [
        ExceptionTestConfiguration(
            ValueError, "raise ValueError('message')", "message"
        ),
        ExceptionTestConfiguration(KeyError, "raise KeyError('message')", "u'message'"),
        ExceptionTestConfiguration(
            OSError, "raise OSError(2, 'message')", "[Errno 2] message"
        ),
        ExceptionTestConfiguration(IOError, "raise IOError('message')", "message"),
        ExceptionTestConfiguration(
            SyntaxError,
            "in valid syntax",
            expected_inner_exception_message="invalid syntax (<string>, line 1)",
            expected_traceback_message=(
                '  File "<string>", line 1\n'
                "    in valid syntax\n"
                "     ^\n"
                "ReraisedSyntaxError: invalid syntax\n"
            ),
        ),
        ExceptionTestConfiguration(
            UnicodeDecodeError,
            "u'£'.encode('utf-8').decode('ascii')",
            "'ascii' codec can't decode byte 0xc2 in position 0: ordinal not in range(128)",
        ),
        ExceptionTestConfiguration(
            UnicodeEncodeError,
            "u'£'.encode('ascii')",
            "'ascii' codec can't encode character u'\\xa3' in position 0: ordinal not in range(128)",
        ),
        ExceptionTestConfiguration(
            AttributeError, "raise AttributeError('message')", "message"
        ),
        ExceptionTestConfiguration(OSError, "raise OSError()"),
        ExceptionTestConfiguration(OSError, "raise OSError(1)", "1"),
        ExceptionTestConfiguration(
            OSError, "raise OSError(2, '£ message')", "[Errno 2] £ message"
        ),
        ExceptionTestConfiguration(
            IOError,
            "raise IOError('исключение')",
            "исключение",
            expected_traceback_message="IOError: <unprintable IOError object>\n",
        ),
        ExceptionTestConfiguration(
            OSError,
            "import locale; raise OSError(2, '£ message'.encode(locale.getpreferredencoding()))",
            "[Errno 2] £ message",
        ),
        ExceptionTestConfiguration(
            OSError, "raise OSError(2, b'£ message')", "[Errno 2] £ message"
        ),
        ExceptionTestConfiguration(
            IOError,
            "raise IOError(b'£ message')",
            "£ message",
            expected_traceback_message="IOError: <unprintable IOError object>\n",
        ),
        ExceptionTestConfiguration(
            Exception, "raise Exception(b'£ message')", "£ message"
        ),
    ],
    ids=[
        "ValueError",
        "KeyError",
        "OSError",
        "IOError",
        "SyntaxError",
        "UnicodeDecodeError",
        "UnicodeEncodeError",
        "AttributeError",
        "OSError - empty",
        "OSError - ErrorNo, empty message",
        "OSError - ErrorNo, unicode message",
        "IOError - unicode message",
        "OSError - bytes message in UTF-8",
        "OSError - bytes message in locale.getpreferredencoding()",
        "IOError - bytes message",
        "Exception - bytes message",
    ],
)


@parametrized_exceptions
def testReraiseKeepsTraceback(exception_configuration):
    with pytest.raises(exception_configuration.exception_type) as e:
        exception_configuration.RaiseExceptionUsingReraise()

    # Reraise() should not appear in the traceback
    traceback_filenames = []
    for traceback_entry in e.traceback:
        try:
            traceback_filenames.append(traceback_entry.path.basename)
        except AttributeError:
            # traceback_entry.path will be a string when the code was run using "exec compile('code')".
            # However this does not hold true if 'code' contains a SyntaxError, which in that case
            # 'traceback_entry.path' will be a LocalPath
            traceback_filenames.append(traceback_entry.path)

    crash_entry = e.traceback.getcrashentry()
    if isinstance(crash_entry.path, six.string_types):
        assert traceback_filenames[0] == b"test_reraise.py"
        assert traceback_filenames[-1] == b"<string>"
    else:
        assert traceback_filenames[0] == b"test_reraise.py"
        assert traceback_filenames[-1] == b"test_reraise.py"


@parametrized_exceptions
def testReraiseAddsMessagesCorrectly(exception_configuration):
    with pytest.raises(exception_configuration.exception_type) as e:
        exception_configuration.RaiseExceptionUsingReraise()

    assert isinstance(e.value, exception_configuration.exception_type)
    assert (
        ExceptionToUnicode(e.value)
        == exception_configuration.GetExpectedExceptionMessage()
    )

    traceback_message = traceback.format_exception_only(type(e.value), e.value)
    traceback_message = b"".join(traceback_message)

    if traceback_message != exception_configuration.GetExpectedTracebackMessage(
        e.value, True
    ):
        assert traceback_message == exception_configuration.GetExpectedTracebackMessage(
            e.value, False
        )


@parametrized_exceptions
def testPickle(exception_configuration):
    try:
        exception_configuration.RaiseExceptionUsingReraise()
    except exception_configuration.exception_type as reraised_exception:
        import cPickle

        dumped_exception = cPickle.dumps(reraised_exception)
        pickled_exception = cPickle.loads(dumped_exception)
        assert ExceptionToUnicode(pickled_exception) == ExceptionToUnicode(
            reraised_exception
        )
        assert ExceptionToUnicode(pickled_exception) != ""
        assert ExceptionToUnicode(reraised_exception) != ""



from __future__ import unicode_literals

import pytest

from barril._foundation.singleton import (
    PushPopSingletonError,
    Singleton,
    SingletonAlreadySetError,
    SingletonNotSetError,
)


def CheckCurrentSingleton(singleton_class, value):
    singleton = singleton_class.GetSingleton()
    assert singleton.value == value


def testSingleton():
    class MySingleton(Singleton):
        def __init__(self, value):
            self.value = value

        @classmethod
        def CreateDefaultSingleton(cls):
            return MySingleton(value=0)

    # Default singleton (created automatically and also put in the stack)
    CheckCurrentSingleton(MySingleton, 0)
    default_singleton = MySingleton.GetSingleton()
    default_singleton.value = 10

    # SetSingleton must be called only when there is no singleton set. In this case,
    # GetSingleton already set the singleton.
    with pytest.raises(SingletonAlreadySetError):
        MySingleton.SetSingleton(MySingleton(value=999))
    CheckCurrentSingleton(MySingleton, 10)

    # push a new instance and test it
    MySingleton.PushSingleton(MySingleton(2000))
    CheckCurrentSingleton(MySingleton, 2000)

    # Calling SetSingleton after using Push/Pop is an error: we do this so that
    # in tests we know someone is doing a SetSingleton when they shouldn't
    with pytest.raises(PushPopSingletonError):
        MySingleton.SetSingleton(MySingleton(value=10))

    # pop, returns to the initial
    MySingleton.PopSingleton()
    CheckCurrentSingleton(MySingleton, 10)

    # SetSingleton given SingletonAlreadySet when outside Push/Pop
    with pytest.raises(SingletonAlreadySetError):
        MySingleton.SetSingleton(MySingleton(value=999))
    CheckCurrentSingleton(MySingleton, 10)

    # The singleton set with "SetSingleton" or created automatically by "GetSingleton" is not
    # part of the stack
    with pytest.raises(PushPopSingletonError):
        MySingleton.PopSingleton()


def testSetSingleton():
    class MySingleton(Singleton):
        def __init__(self, value=None):
            self.value = value

    assert not MySingleton.HasSingleton()

    MySingleton.SetSingleton(MySingleton(value=999))
    assert MySingleton.HasSingleton()
    CheckCurrentSingleton(MySingleton, 999)

    with pytest.raises(SingletonAlreadySetError):
        MySingleton.SetSingleton(MySingleton(value=999))

    MySingleton.ClearSingleton()
    assert not MySingleton.HasSingleton()

    with pytest.raises(SingletonNotSetError):
        MySingleton.ClearSingleton()


def testPushPop():
    class MySingleton(Singleton):
        def __init__(self, value=None):
            self.value = value

    MySingleton.PushSingleton()

    assert MySingleton.GetStackCount() == 1

    with pytest.raises(PushPopSingletonError):
        MySingleton.ClearSingleton()

    MySingleton.PushSingleton()
    assert MySingleton.GetStackCount() == 2

    MySingleton.PopSingleton()
    assert MySingleton.GetStackCount() == 1

    MySingleton.PopSingleton()
    assert MySingleton.GetStackCount() == 0

    with pytest.raises(PushPopSingletonError):
        MySingleton.PopSingleton()


def testGetSingletonThreadSafe(mocker):
    from threading import Event, Thread

    class MySingleton(Singleton):
        @classmethod
        def SlowConstructor(cls, event):
            event.wait(1)
            return MySingleton()

    thrlist = [Thread(target=MySingleton.GetSingleton) for _ in range(3)]
    create_singleton_mock = mocker.patch.object(MySingleton, "CreateDefaultSingleton")

    event = Event()
    create_singleton_mock.side_effect = lambda: MySingleton.SlowConstructor(event)
    for thread in thrlist:
        thread.start()
    event.set()
    for thread in thrlist:
        thread.join()
    assert create_singleton_mock.call_count == 1



from __future__ import absolute_import, unicode_literals

import six

from barril.units._quantity import _Quantity
from barril.units.unit_database import UnitDatabase

from ._quantity import ObtainQuantity

__all__ = [str("AbstractValueWithQuantityObject")]  # pylint: disable=invalid-all-object


class AbstractValueWithQuantityObject(object):
    """
    This is an abstract class that provides a default implementation for having a class
    that has a value associated with a quantity.

    Subclasses should only provide a method for getting and setting the value, as it may be a
    numarray object, list of floats, etc (so, the user may want to call it GetValue, GetValues,
    GetNumarray, ...).

    .. see:: L{Scalar} for an implementation example
    """

    def __init__(self, category, value=None, unit=None):

        unit_database = UnitDatabase.GetSingleton()
        if category.__class__ == _Quantity:
            quantity = category

            assert unit is None, "If quantity is given, the unit must not!"

            if value is None:
                value = self._GetDefaultValue(quantity.GetCategoryInfo())

        else:
            if category.__class__ != six.text_type:
                # Support for creating a scalar as:
                # Scalar(10, 'm')
                # Scalar(10, 'm', 'length')
                value, unit, category = category, value, unit

            if value is None or unit is None:
                if value is None:
                    category_info = unit_database.GetCategoryInfo(category)
                    value = self._GetDefaultValue(category_info, unit)
                    if unit is None:
                        unit = category_info.default_unit
                        assert unit is not None
                else:
                    # Note: if the category is given with a value, the unit MUST also be given.
                    # (because if later we change the default unit, the original category would be wrong).
                    assert (
                        unit is not None
                    ), "If category and value are given, the unit must be specified too."

            assert type(unit) is not bytes
            assert type(category) is not bytes
            quantity = ObtainQuantity(unit, category)

        self._InternalCreateWithQuantity(quantity, value, unit_database)

    def IsValid(self):
        """
        :rtype: bool
        :returns:
            Checks if current state is valid.
            Should be implemented by each derived implementation.
        """
        if self.GetQuantity().IsDerived():
            return True
        try:
            self.CheckValidity()
        except ValueError:
            return False
        return True

    def HasCategory(self):
        """
        Returns whether this instance has any associated category

        :rtype: bool
        :returns:
            If there's some category composing this object
        """
        return len(self._quantity.GetComposingCategories()) > 0

    # UnitDatabase ---------------------------------------------------------------------------------
    def GetUnitDatabase(self):
        """
        :rtype: UnitDatabase
        :returns:
            The unit database that was used in the creation of this object.
        """
        return self._quantity.GetUnitDatabase()

    # Quantity -------------------------------------------------------------------------------------
    def GetQuantity(self):
        """
        :rtype: Quantity
        :returns:
            The Quantity that is associated with this object.
        """
        return self._quantity

    # Category -------------------------------------------------------------------------------------
    def GetCategory(self):
        return self._quantity.GetCategory()

    category = property(GetCategory)

    # QuantityType ---------------------------------------------------------------------------------
    def GetQuantityType(self):
        return self._quantity.GetQuantityType()

    quantity_type = property(GetQuantityType)

    # Unit -----------------------------------------------------------------------------------------
    def GetUnit(self):
        return self._quantity.GetUnit()

    unit = property(GetUnit)

    # UnitDatabase Shortcuts -----------------------------------------------------------------------
    def GetUnitName(self):
        """
        :rtype: unicode
        :returns:
            Returns a user-friendly name for the given unit (i.e.: 'm' would me 'meters')
        """
        return self._quantity.GetUnitName()

    def GetValidUnits(self):
        """
        :rtype: list(unicode)
        :returns:
            Returns a list with all the valid units for the category for this object + the current
            unit if it's not in the list of valid units (because it may be using a unit valid for
            the quantity type even if it doesn't match a unit in the category).
        """
        valid_units = self.GetUnitDatabase().GetValidUnits(self.GetCategory())
        current_unit = self.GetQuantity().GetUnit()
        if current_unit not in valid_units:
            valid_units.append(current_unit)
        return valid_units

    # ===============================================================================================
    # __Stub
    # ===============================================================================================
    class __Stub(object):
        """
        Helper class for used in CreateWithQuantity.
        """

    @classmethod
    def CreateWithQuantity(cls, quantity, *args, **kwargs):
        """
        This is a secondary interface for creating the object with an existing quantity.

        :param Quantity quantity:
            The quantity for this object

        @param args and kwargs: Those are dependent on the actual class -- this parameters
            are passed directly to _InternalCreateWithQuantity.
        """

        stub = cls.__Stub()
        stub.__class__ = cls
        stub._InternalCreateWithQuantity(quantity, *args, **kwargs)
        return stub

    # Copy -----------------------------------------------------------------------------------------
    def Copy(self):
        """
        :rtype: Scalar
        :returns:
            Returns self, since Scalar is immutable
        """
        return self

    def CreateCopyInstance(self):
        """
        :rtype: Scalar
        :returns:
            Returns a new scalar that's a copy of this scalar.
        """
        return self

    def __copy__(self):
        """
        Copy protocol.
        """
        return self.Copy()

    def __deepcopy__(self, memo):
        """
        Copy protocol.
        """
        return self.Copy()

    def CreateCopy(self, value=None, unit=None, category=None, **kwargs):
        """
        :rtype: Scalar
        :returns:
            Returns a new scalar that's a copy of this scalar.
        """
        try:
            if value is None:
                value = self.GetAbstractValue(unit)

            if unit is None and category is None:
                return self.CreateWithQuantity(self._quantity, value=value, **kwargs)

            elif category is not None:
                if unit is None:
                    raise TypeError(
                        "If category is given, the unit must be specified too."
                    )

                return self.CreateWithQuantity(
                    ObtainQuantity(unit, category), value=value, **kwargs
                )

            elif unit is not None:
                if self._quantity.GetCategory():
                    return self.CreateWithQuantity(
                        ObtainQuantity(unit, self._quantity.GetCategory()),
                        value=value,
                        **kwargs
                    )
                else:
                    # Handling empty quantity
                    return self.CreateWithQuantity(
                        ObtainQuantity(unit), value=value, **kwargs
                    )

            else:
                raise RuntimeError("Execution should never get here!")

        except TypeError as e:
            raise TypeError(
                "Error creating new instance of %s: %s\n"
                "(Should be overridden in '%s' class if it takes parameters in __init__)"
                % (self.__class__.__name__, e, self.__class__.__name__)
            )

    def __ne__(self, other):
        return not self == other

    def __hash__(self, *args, **kwargs):
        raise NotImplementedError(
            "Objects with a quantity are not hashable (as they're usually mutable)."
        )

    # Format ---------------------------------------------------------------------------------------

    FORMATTED_SUFFIX_FORMAT = " [%s]"

    @classmethod
    def GetFormattedSuffixFormat(cls):
        """
        Returns the formatted suffix for the unit.

        :rtype: unicode
        :returns:
            The formatted suffix
        """
        return cls.FORMATTED_SUFFIX_FORMAT

    @classmethod
    def SetFormattedSuffixFormat(cls, pattern):
        """
        Sets the format for the formatted text suffix, which may include the unit (Use "%s" to
        place the unit.).

        :param unicode pattern:
            A format-like string containing one C{%s} format code
        """
        try:
            pattern % "unit"
        except TypeError as e:
            from barril._foundation.reraise import Reraise

            Reraise(
                e,
                "Incompatible pattern for Scalar suffix. Expected a format for a unicode value.",
            )
        cls.FORMATTED_SUFFIX_FORMAT = pattern

    def GetFormattedSuffix(self, unit=None):
        """
        Returns the suffix for the formatted string using the current unit.

        :rtype: unicode
        :returns:
            The suffix
        """
        if unit is None:
            unit = self.GetUnit()
        return self.FORMATTED_SUFFIX_FORMAT % unit

    # : sentinel used to change the category of the object (see ChangeCategory in subclasses)
    DEFAULT_CATEGORY_VALUE = object()
    # : sentinel used to change the category of the object (see ChangeCategory in subclasses)
    CURRENT_VALUE = object()



"""
Extensions to python native types.
"""

from __future__ import unicode_literals

import six

from barril._foundation.klass import IsInstance

_TRUE_VALUES = ["TRUE", "YES", "1"]
_FALSE_VALUES = ["FALSE", "NO", "0"]
_TRUE_FALSE_VALUES = _TRUE_VALUES + _FALSE_VALUES
_KNOWN_NUMBER_TYPES = None


def _GetKnownNumberTypes():
    """
    Dynamically obtain the tuple with the types considered number, including numpy.number if
    possible.

    This code replaces an old implementation with "code replacement". Not checked if we have any
    performance penalties.
    """
    result = {float, complex}
    result.update(set(six.integer_types))
    try:
        import numpy
    except ImportError:
        pass
    else:
        result.add(numpy.number)
    return tuple(result)


# ===================================================================================================
# Boolean
# ===================================================================================================
def Boolean(text):
    """
    :param str text:
        A text semantically representing a boolean value.

    :rtype: bool
    :returns:
        Returns a boolean represented by the given text.
    """
    text_upper = text.upper()
    if text_upper not in _TRUE_FALSE_VALUES:
        raise ValueError(
            "The value does not match any known value (case insensitive): %s (%s)"
            % (text, _TRUE_FALSE_VALUES)
        )
    return text_upper in _TRUE_VALUES


# ===================================================================================================
# MakeTuple
# ===================================================================================================
def MakeTuple(object_):
    """
    Returns the given object as a tuple, if it is not, creates one with it inside.

    @param: Any object or tuple
        The object to tupleIZE
    """
    if isinstance(object_, tuple):
        return object_
    else:
        return (object_,)


# ===================================================================================================
# CheckType
# ===================================================================================================
def CheckType(object_, type_, message=None):
    """
    Check if the given object is of the given type, raising a descriptive "TypeError" if it is
    not.

    :type object_: Any object
    :param object_:
        The object to check the type

    :type type_: type, tuple of types, type name, tuple of type names
    :param type_:
        The type or types to check.
        This can be a actual type or the name of the type.
        This can be one type or a list of types.

    :param str message:
        The error message shown if the check does not pass.
        By default, generates the following message:

            CheckType: Expecting <expected type(s)>, got <object type>: <object>

    @raise: TypeError:
        Raises type-error if the object does not matches the given type(s).
    """
    result = IsInstance(object_, type_)
    if not result:
        # 001) The list of the types names.
        type_names = [x if isinstance(x, str) else x.__name__ for x in MakeTuple(type_)]
        type_names = '" or "'.join(type_names)

        # 002) Build the error message
        exception_message = 'CheckType: Expecting "%s", got "%s": %s' % (
            type_names,
            object_.__class__.__name__,
            repr(object_),
        )

        # 003) Appends the user message at the end
        if message:
            exception_message += "\n%s" % message

        # 004) Raises the exception
        raise TypeError(exception_message)

    return result


# ===================================================================================================
# CheckFormatString
# ===================================================================================================
def CheckFormatString(pattern, *arguments):
    """
    Checks if the given format string (for instance, "%.g") is valid for the given arguments.
    :param pattern: a string in "%" format
    :param arguments: the arguments to check against
    :raises ValueError:
        if the format is invalid.
    """
    try:
        pattern % arguments
    except (TypeError, ValueError):
        raise ValueError("%r is not a valid format string." % pattern)


# ===================================================================================================
# _IsNumber
# ===================================================================================================
def _IsNumber(v):
    """
    Actual function code for IsNumber.

    Checks if the given value is a number

    @return bool
        True if the given value is a number, False otherwise
    """
    return isinstance(v, _KNOWN_NUMBER_TYPES)


# ===================================================================================================
# IsNumber
# ===================================================================================================
def IsNumber(v):
    """
    Checks if the given value is a number

    :return bool:
        True if the given value is a number, False otherwise

    .. note:: This function will replace it's implementation to a lighter code, but first it must
        define which types are known as a number.
        The code replacement is made to avoid the call to import and listing the know numeric types.
    """
    # There are cases were the numpy will not be available (for example when the aasimar is building
    # the environment the numpy is not available yet). Delegate this import to the IsNumber would
    # cause a severe performance impact. So we will attempt to import the numpy, but if the lib is
    # not available let us move on the know number types.
    global _KNOWN_NUMBER_TYPES
    _KNOWN_NUMBER_TYPES = _GetKnownNumberTypes()

    # Replaces this function by an optimized version _IsNumber.
    IsNumber.__code__ = _IsNumber.__code__

    return _IsNumber(v)


# ===================================================================================================
# CheckIsNumber
# ===================================================================================================
def CheckIsNumber(v):
    if not IsNumber(v):
        raise TypeError("Expecting a number. Received:%s (%s)" % (v, type(v)))
    return True


# ===================================================================================================
# IsBasicType
# ===================================================================================================
def IsBasicType(value, accept_compound=False, additional=None):
    """
    :param object value:
        The value we want to check

    :param bool accept_compound:
        Whether lists, tuples, sets and dicts should be considered basic types

    :type additional: class or tuple(classes)
    :param additional:
        Any classes in this additional list will also be considered as basic types.

    :rtype: bool
    :returns:
        True if the passed value is from a basic python type
    """
    if (
        isinstance(value, _ACCEPTED_BASIC_TYPES)
        or value is None
        or (additional and IsInstance(value, additional))
    ):
        return True

    if accept_compound:
        if isinstance(value, dict):
            for key, val in six.iteritems(value):
                if not IsBasicType(key, accept_compound, additional) or not IsBasicType(
                    val, accept_compound, additional
                ):
                    return False
            return True

        if isinstance(value, (tuple, list, set, frozenset)):
            for val in value:
                if not IsBasicType(val, accept_compound, additional):
                    return False
            return True

    return False


_ACCEPTED_BASIC_TYPES = tuple(
    list(six.string_types) + list(six.integer_types) + [bytes, float, bool, complex]
)


# ===================================================================================================
# CheckBasicType
# ===================================================================================================
def CheckBasicType(value, accept_compound=False, additional=None):
    """
    .. see:: IsBasicType for parameters descriptions.

    :rtype: True
    :returns:
        True if the passed value is from a basic python type

    :raises TypeError:
        if the value passed is not a basic type
    """
    if not IsBasicType(value, accept_compound, additional):
        raise TypeError(
            "Expecting a basic type. Received:%s (%s)" % (value, type(value))
        )
    return True


# ===================================================================================================
# CheckEnum
# ===================================================================================================
def CheckEnum(value, enum_values):
    """
    Checks if the given value belongs to the given enum. This function is meant to replace code like
    this:

    ENUM_VALUE_1 = 'value1'
    ENUM_VALUE_2 = 'value2'
    ENUM_VALUES = set([ENUM_VALUE_1, ENUM_VALUE_2])

    def Foo(value):
        if value not in ENUM_VALUES:
            raise ValueError(...)

    :param object value:
        The value to test membership in the enum

    :type values: sequence of objects
    :param values:
        The values that are acceptable for this enum.

    :raises ValueError:
        if the given value does not belong to the enum.
    """
    if value not in enum_values:
        msg = "The value %r is not valid for the expected num: %s"
        raise ValueError(msg % (value, list(enum_values)))


# ===================================================================================================
# Intersection
# ===================================================================================================
def Intersection(*sequences):
    """
    Return the intersection of all the elements in the given sequences, ie, the items common to all
    the sequences.

    :type sequences: a list of sequences.
    :param sequences:
    :rtype: a set() with the intersection.
    """
    if not sequences:
        return set()
    result = set(sequences[0])
    for seq in sequences[1:]:
        result.intersection_update(seq)
    return result


# ===================================================================================================
# OrderedIntersection
# ===================================================================================================
def OrderedIntersection(*sequences):
    """
    Like Intersection, but the returned sequence is in the order of the first one.
    """
    intersection = Intersection(*sequences)
    if not intersection:
        return []

    return [x for x in sequences[0] if x in intersection]


# ===================================================================================================
# AsList
# ===================================================================================================
def AsList(arg):
    """Returns the given argument as a list; if already a list, return it unchanged, otherwise
    return a list with the arg as only element.
    """
    if isinstance(arg, (list)):
        return arg

    if isinstance(arg, (tuple, set)):
        return list(arg)

    return [arg]


# ===================================================================================================
# Flatten
# ===================================================================================================
def Flatten(iterable, skip_types=None):
    """
    :rtype: list
    :returns:
        A list with the elements of the initial iterable flattened.
    """
    return list(IterFlattened(iterable, skip_types))


# ===================================================================================================
# IterFlattened
# ===================================================================================================
def IterFlattened(iterable, skip_types=None):
    """
    Flattens recursively the passed iterable with subsequences into a flat iterator.

    Warning: This method also flattens tuples

    :type iterable: an iterable object, ie, an object that iter() can handle
    :param iterable:
        The iterable to be flattened.

    :param list(type) skip_types:
        These types won't be flattened if they happen to be iterable.

    :rtype: iterator
    :returns: Iterator over the flattened given iterable.
    """
    if skip_types is None:
        skip_types = []

    for s in six.string_types:
        if s not in skip_types:
            # Exceptional case: we want to treat strings as elements!
            skip_types.append(s)

    skip_types_tuple = tuple(skip_types)

    for element in iterable:
        element_iter = (
            None
        )  # will be None if element is not iterable, or hold an iterator otherwise

        if not isinstance(element, skip_types_tuple):
            try:
                element_iter = iter(element)
            except TypeError:
                pass

        if element_iter is None:
            yield element
        else:
            for x in IterFlattened(element_iter, skip_types):
                yield x


# ===================================================================================================
# MergeDictsRecursively
# ===================================================================================================
def MergeDictsRecursively(original_dict, merging_dict):
    """
    Merges two dictionaries by iterating over both of their keys and returning the merge
    of each dict contained within both dictionaries.
    The outer dict is also merged.

    ATTENTION: The :param(merging_dict) is modified in the process!

    :param dict original_dict
    :param dict merging_dict
    """
    items = iter(original_dict.items())

    for key, value in items:
        try:
            other_value = merging_dict[key]
            MergeDictsRecursively(value, other_value)
            del merging_dict[key]
        except KeyError:
            continue
        except TypeError:
            continue
        except AttributeError:
            continue

    try:
        original_dict.update(merging_dict)
    except ValueError:
        raise TypeError(
            'Wrong types passed. Expecting two dictionaries, got: "%s" and "%s"'
            % (type(original_dict).__name__, type(merging_dict).__name__)
        )

    return original_dict


# ===================================================================================================
# ListDuplicates
# ===================================================================================================
def ListDuplicates(iterable):
    """
    Given a sequence, returns a list containing all the items in 'iterable' that appear more than
    once. E.g. ListDuplicates([1,2,3,1]) returns [1].

    :type iterable: an iterable object, ie, an object that iter() can handle.
    """
    seen = set()
    seen_add = seen.add
    # adds all elements it doesn't know yet to seen and all other to seen_twice
    seen_twice = set(x for x in iterable if x in seen or seen_add(x))
    # turn the set into a list (as requested)
    return list(seen_twice)


# ===================================================================================================
# StructMap
# ===================================================================================================
def StructMap(obj, func, conditional=lambda x: True):
    """
    Maps an object recursively for dict/tuple/list types.

    :param callable func:
        Called for each non-container value IF conditional returns True.

    :param callable conditional:
        Filters which values to execute func.

    :return object:
        Returns a copy of the object with all non-container types mapped via func.

    Example:
        Converts all "bytes" instances to "str" inside nested container objects

        struct = {
            'name' : 'Alpha',
            'numbers' : (1,2,3),
            'surnames' : ('Bravo', 'Charlie'),
        }

        struct = StructMap(
            struct,
            lambda x: x.decode('UTF-8'),
            lambda x: isinstance(x, str)
        )

        struct = {
            u'name' : u'Alpha',
            u'numbers' : (1,2,3),
            u'surnames' : (u'Bravo', u'Charlie'),
        }
    """
    if isinstance(obj, dict):
        return {
            StructMap(i, func, conditional): StructMap(j, func, conditional)
            for i, j in obj.items()
        }
    if isinstance(obj, tuple):
        return tuple(StructMap(i, func, conditional) for i in obj)
    if isinstance(obj, list):
        return [StructMap(i, func, conditional) for i in obj]
    if conditional(obj):
        obj = func(obj)
    return obj



# coding: UTF-8
from __future__ import unicode_literals

import locale
import sys

import pytest
import six

from barril._foundation.exceptions import ExceptionToUnicode

skip_py3_obsolete_unicode = pytest.mark.skipif(
    condition=not six.PY2, reason="Python 3 turned unicode conversions obsolete"
)


@skip_py3_obsolete_unicode
def testExceptionToUnicode(
    exception_message, lpe_exception_message, fse_exception_message
):
    assert ExceptionToUnicode(Exception(exception_message)) == exception_message
    assert ExceptionToUnicode(Exception(fse_exception_message)) == exception_message
    assert ExceptionToUnicode(Exception(lpe_exception_message)) == exception_message


@skip_py3_obsolete_unicode
def testExceptionToUnicodeOSError(
    exception_message, lpe_exception_message, fse_exception_message
):
    assert (
        ExceptionToUnicode(OSError(2, exception_message))
        == "[Errno 2] " + exception_message
    )
    assert (
        ExceptionToUnicode(OSError(2, fse_exception_message))
        == "[Errno 2] " + exception_message
    )
    assert (
        ExceptionToUnicode(OSError(2, lpe_exception_message))
        == "[Errno 2] " + exception_message
    )


@skip_py3_obsolete_unicode
def testExceptionToUnicodeIOError(
    exception_message, lpe_exception_message, fse_exception_message
):
    # IOError is really stupid, unicode(IOError('á')) actually raises UnicodeEncodeError
    # (not UnicodeDecodeError!)
    assert ExceptionToUnicode(IOError(exception_message)) == exception_message
    assert ExceptionToUnicode(IOError(fse_exception_message)) == exception_message
    assert ExceptionToUnicode(IOError(lpe_exception_message)) == exception_message


@skip_py3_obsolete_unicode
def testExceptionToUnicodeCustomException(
    exception_message, lpe_exception_message, fse_exception_message
):
    class MyException(Exception):
        def __unicode__(self):
            return "hardcoded unicode repr"

    assert (
        ExceptionToUnicode(MyException(exception_message)) == "hardcoded unicode repr"
    )


@skip_py3_obsolete_unicode
def testExceptionToUnicodeBadEncoding(
    exception_message, lpe_exception_message, fse_exception_message
):
    assert (
        ExceptionToUnicode(Exception(b"random \x90\xa1\xa2"))
        == "random \ufffd\ufffd\ufffd"
    )


@skip_py3_obsolete_unicode
def testExceptionToUnicodeUTF8():
    assert ExceptionToUnicode(Exception("Ação".encode("utf-8"))) == "Ação"


def testExceptionToUnicodeWithReraise():
    first_caption = "first"
    second_caption = "second"
    third_caption = "third"

    def FunctionWithReraise():
        from barril._foundation.reraise import Reraise

        try:
            raise RuntimeError(first_caption)
        except RuntimeError as e:
            try:
                Reraise(e, second_caption)
            except RuntimeError as e2:
                Reraise(e2, third_caption)

    with pytest.raises(RuntimeError) as ex_info:
        FunctionWithReraise()

    exception_message = ExceptionToUnicode(ex_info.value)
    for caption in [first_caption, second_caption, third_caption]:
        assert caption in exception_message


@pytest.mark.skipif(condition=six.PY2, reason="Python 3 exception context")
def testExceptionToUnicodeWithRaiseContext():
    """Ensure ExceptionToUnicode builds a message with all context/causes of an exception"""
    with pytest.raises(RuntimeError) as exc_info:

        def Original():
            raise RuntimeError("Original Error")

        try:
            Original()
        except RuntimeError:
            raise RuntimeError("While handling error")

    assert ExceptionToUnicode(exc_info.value) == "While handling error\nOriginal Error"


# ===================================================================================================
# Fixtures
# ===================================================================================================
@pytest.fixture
def exception_message():
    exception_message = "кодирование"
    # Use another message if this machine's locale does not support cyrilic
    try:
        exception_message.encode(locale.getpreferredencoding())
    except UnicodeEncodeError:
        exception_message = "látïn-1"

    return exception_message


@pytest.fixture
def lpe_exception_message(exception_message):
    return exception_message.encode(locale.getpreferredencoding())


@pytest.fixture
def fse_exception_message(exception_message):
    return exception_message.encode(sys.getfilesystemencoding())



from __future__ import absolute_import, division, unicode_literals

import copy

import pytest
import six

from barril.basic.fraction import Fraction


def testBasicUsage():
    f = Fraction(5, 3)
    assert tuple(f) == (5, 3)
    assert len(f) == 2
    assert f[0] == 5
    assert f[1] == 3
    assert f.numerator == 5
    assert f.denominator == 3
    f.numerator = 10
    assert tuple(f) == (10, 3)
    f.denominator = 4
    assert tuple(f) == (5, 2)


def testStr():
    assert six.text_type(Fraction(5, 3)) == "5/3"
    assert repr(Fraction(5, 3)) == "Fraction(5, 3)"

    assert six.text_type(Fraction(3 / 1000, 4)) == "3/4000"


def testReduce():
    f = Fraction(5, 3)
    f[0] = 15
    assert tuple(f) == (5, 1)


def testOperations():
    # float operator
    assert float(Fraction(5, 3)) == 5.0 / 3.0

    # sum
    assert Fraction(5, 3) + Fraction(2, 3) == Fraction(7, 3)
    assert Fraction(5, 3) + 5.0 == Fraction(20, 3)
    assert 5.0 + Fraction(5, 3) == Fraction(20, 3)

    # sub
    assert Fraction(5, 3) - Fraction(2, 3) == Fraction(3, 3)
    assert Fraction(5, 3) - 1.0 == Fraction(2, 3)
    assert 1.0 - Fraction(5, 3) == Fraction(-2, 3)

    # mul
    assert Fraction(5, 3) * Fraction(4, 3) == Fraction(20, 9)
    assert 3 * Fraction(5, 3) == Fraction(15, 3)
    assert Fraction(5, 3) * 3 == Fraction(15, 3)

    # div
    assert Fraction(5, 3) / Fraction(5, 3) == Fraction(15, 15)
    assert Fraction(5, 3) / 3 == Fraction(5, 9)
    assert 3 / Fraction(5, 3) == Fraction(9, 5)

    # inv
    assert Fraction(5, 3).inv() == Fraction(3, 5)

    # neg
    assert -Fraction(5, 3) == Fraction(-5, 3)


def testCopy():
    f = Fraction(5, 3)
    cf = copy.copy(f)
    assert f == cf
    assert not f != cf
    assert f is not cf

    assert f != Fraction(3, 5)
    assert not f == Fraction(3, 5)


def testStrFormat():
    """
    Scalar field behavior. In this test we make sure that the
    Fraction.__str__ method calls FormatFloat, which handles
    the locale properly.
    """
    import barril.basic.format_float

    # By default, the numbers are formatted using "%g"
    f = Fraction(5, 3)
    assert six.text_type(f), "5/3"

    # Test the use of FormatFloat | 5.6 is 28/5 | 5.6/3 is 28/15
    f = Fraction(5.6, 3)
    assert six.text_type(f) == "28/15"

    original_format_float = barril.basic.format_float.FormatFloat
    barril.basic.format_float.FormatFloat = lambda x, y: "X%.2fX" % y
    try:
        assert six.text_type(f) == "X28.00X/X15.00X"
    finally:
        barril.basic.format_float.FormatFloat = original_format_float

    assert six.text_type(f) == "28/15"


@pytest.mark.skipif(
    condition=not six.PY2, reason="Long type is no longer available in Python 3"
)
def testLong():
    f = Fraction(20000000000000000, None)
    assert tuple(f) == (20000000000000000, 1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

import io
from setuptools import find_packages, setup

with io.open("README.rst", encoding="UTF-8") as readme_file:
    readme = readme_file.read()

with io.open("CHANGELOG.rst", encoding="UTF-8") as changelog_file:
    history = changelog_file.read()

requirements = ["attrs>=18.1.0", "numpy>=1.11.0"]
extras_require = {
    ':python_version == "2.7"': ["ruamel.ordereddict>=0.4.6"],
    "docs": ["sphinx >= 1.4", "sphinx_rtd_theme", "sphinx-autodoc-typehints"],
    "testing": ["codecov", "pytest", "pytest-cov", "pytest-mock", "pre-commit", "tox"],
}

setup(
    author="ESSS",
    author_email="foss@esss.co",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    description="Python package to manage units for physical quantities",
    extras_require=extras_require,
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*",
    keywords="barril",
    name="barril",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    url="https://github.com/ESSS/barril",
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    zip_safe=False,
)



from __future__ import absolute_import, division, unicode_literals

import pytest
from pytest import approx
import six
from six.moves import range, zip

from barril._foundation.odict import odict
from barril import units
from barril.units import Array, InvalidUnitError, ObtainQuantity, Quantity


def testEmptyArray():
    arr = Array.CreateEmptyArray()
    assert not arr.HasCategory()

    arr = arr.CreateCopy(category="temperature", unit="degC")
    assert arr.HasCategory()


def testValues():
    array = units.Array("length", values=[100, 150, 200], unit="m")
    assert array.unit == "m"
    assert array.values == [100, 150, 200]

    array = array.CreateCopy(values=list(range(100)))
    array = array.CreateCopy(unit="km")
    assert array.values == [x / 1000.0 for x in range(100)]
    assert 3.0 / 1000.0 == array[3]

    with pytest.raises(AttributeError):
        array.values = 10


def testArrayWithNaNs(unit_database_len):
    import numpy

    TEST_VALUES = [[numpy.nan, 0, 15, numpy.nan], [numpy.nan] * 5]
    for values in TEST_VALUES:
        units.Array(
            "flow rate", values=values, unit="m3/s"
        )  # check that this does not raise exceptions


def testStr():
    array = units.Array("length", values=[100, 150, 200], unit="m")
    assert six.text_type(array) == "100 150 200 [m]"


def testConvertValuesFromArray():
    def CheckConversion(values):
        _array = units.Array("length", values=values, unit="m")
        for converted, original in zip(_array.GetValues("km"), values):
            assert converted == original / 1000

    from numpy import array

    CheckConversion(array([0.0, 1.123, 2.456, 3.789]))
    CheckConversion(array([0.0, 0.0]))


def testValues2():
    array = units.Array("length", values=[100, 150, 200], unit="m")
    array2 = units.Array("length", values=(100, 150, 200), unit="m")
    assert array == array2


def testEquality():
    a1 = units.Array("length", values=list(range(100)), unit="m")

    assert a1 == units.Array("length", values=list(range(100)), unit="m")
    assert a1 != units.Array("temperature", values=list(range(100)), unit="degC")
    assert a1 != units.Array("length", values=[], unit="m")
    assert a1 != units.Array("length", values=list(range(100)), unit="km")


def testMultiDimensional():
    a1 = units.Array("length", values=[(1000, 1000), (300, 300)], unit="m")
    assert a1.GetValues("km") == [(1, 1), (0.3, 0.3)]

    assert (1000, 1000) == a1[0]
    assert (300, 300) == a1[1]

    a1 = a1.CreateCopy(unit="km")
    assert (1, 1) == a1[0]
    assert (0.3, 0.3) == a1[1]


def testCreateWithQuantity():
    q = ObtainQuantity("m", "length")
    a1 = units.Array(q, values=[(1000, 1000), (300, 300)])
    a2 = a1.CreateCopy(unit="km")

    assert a1.values == [(1000, 1000), (300, 300)]
    assert a2.values == [(1, 1), (0.3, 0.3)]


def testCategoryParameters(unit_database_len, mocker):
    unit_database = unit_database_len
    unit_database.AddUnit(
        "temperature", "Kelvin Degrees", "degK", "%f - 273.5", "%f + 273.5"
    )
    unit_database.AddCategory(
        "my temperature",
        "temperature",
        ["degC", "degK"],
        default_unit="degC",
        default_value=22.3,
        min_value=-10.0,
        max_value=213.4,
        is_min_exclusive=False,
        is_max_exclusive=True,
    )

    a1 = units.Array("my temperature", values=[0, 100], unit="degC")
    assert a1.IsValid()

    a2 = a1.CreateCopy(values=[0, 220])
    assert a2._is_valid is None
    assert a2._validity_exception is None

    mocker.patch.object(a2, "_DoValidateValues", autospec=True)
    a2._DoValidateValues.side_effect = ValueError("my value error")

    assert not a2.IsValid()
    assert a2._is_valid is False
    assert six.text_type(a2._validity_exception) == "my value error"
    assert a2._DoValidateValues.call_count == 1

    with pytest.raises(ValueError):
        a2.CheckValidity()

        # As result is already cached the "_DoValidateValues" will not be invoked.
        assert a2._DoValidateValues.call_count == 1

    a3 = units.Array("my temperature", values=[(-100, 50), (0, 2)], unit="degC")
    assert not a3.IsValid()


def testInvalidUnit():
    a1 = units.Array("length", values=[(1000, 1000), (300, 300)], unit="m")
    try:
        a1 = a1.CreateCopy(values=[(100, 200)], unit="foo")
    except InvalidUnitError:
        pass
    assert a1.values == [(1000, 1000), (300, 300)]
    assert a1.unit == "m"


def testGetValues():
    """
        Tests GetValues method return type when passing a 'unit'
    """
    array = units.Array("temperature", values=[0, 100], unit="degC")
    assert isinstance(array.GetValues(), list)
    assert isinstance(array.GetValues("degC"), list)

    array = units.Array("temperature", values=(0, 100), unit="degC")
    assert isinstance(array.GetValues(), tuple)
    assert isinstance(array.GetValues("degC"), tuple)
    assert array[1] == 100

    array = units.Array("temperature", values=[], unit="degC")
    assert isinstance(array.GetValues("degC"), list)


def testCopy():
    array = units.Array("length", values=[0, 100], unit="m")
    copy = array.Copy()
    assert copy is array

    # test copy of subclasses
    class MyArray(units.Array):
        pass

    array = MyArray("length", values=[0, 100], unit="m")
    copy = array.Copy()
    assert isinstance(copy, MyArray)


def testArrayOperations(unit_database_len_time):
    unit_database = unit_database_len_time

    m = Quantity.CreateDerived(odict([("Table size", ["m", 1])]))
    km_city = Quantity.CreateDerived(odict([("City size", ["km", 1])]))

    s1 = Array.CreateWithQuantity(m, [1])
    s2 = Array.CreateWithQuantity(km_city, [0.01])
    initial1 = s1.GetQuantity().GetComposingUnits()
    initial2 = s2.GetQuantity().GetComposingUnits()
    # Check that they doesn't raise ComposedUnitError
    s1.GetValues()
    s2.GetValues()

    quantity, value = unit_database.Multiply(m, km_city, 1, 0.01)
    assert initial1 == s1.GetQuantity().GetComposingUnits()
    assert initial2 == s2.GetQuantity().GetComposingUnits()
    calculated1 = Array.CreateWithQuantity(quantity, [value])

    array = s1 * s2
    six.text_type(array)  # just to see if it works...
    assert calculated1 == s1 * s2

    quantity, value = unit_database.Sum(m, km_city, 1, 0.01)
    assert Array.CreateWithQuantity(quantity, [value]) == s1 + s2

    quantity, value = unit_database.Subtract(m, km_city, 1, 0.01)
    assert Array.CreateWithQuantity(quantity, [value]) == s1 - s2


def testDivision(unit_database_len_time):
    unit_database = unit_database_len_time
    m = Quantity.CreateDerived(odict([("Table size", ["m", 1])]))
    km_city = Quantity.CreateDerived(odict([("City size", ["km", 1])]))
    quantity, value = unit_database.Divide(m, km_city, 1, 0.01)
    calculated1 = Array.CreateWithQuantity(quantity, [value])
    s1 = Array.CreateWithQuantity(m, [1])
    s2 = Array.CreateWithQuantity(km_city, [0.01])
    assert calculated1 == s1 / s2


def testNumberInteractions(unit_database_len_time):
    import numpy

    m = Quantity.CreateDerived(odict([("Table size", ["m", 1])]))
    s1 = Array.CreateWithQuantity(m, list(range(10)))
    s2 = Array.CreateWithQuantity(m, [x + x for x in range(10)])
    assert s1 == 0 + s1
    assert s1 == s1 + 0
    assert s2 == s1 + s1

    num_arr = Array.CreateWithQuantity(m, numpy.array([x for x in range(10)]))
    sum_array = num_arr + s1
    assert isinstance(sum_array.values, numpy.ndarray)

    sum_array2 = num_arr + numpy.array([x for x in range(10)])
    assert isinstance(sum_array2.values, numpy.ndarray)

    tup_arr = Array.CreateWithQuantity(m, tuple([x for x in range(10)]))
    tup_arr = tup_arr + 1
    assert isinstance(tup_arr.values, tuple)


def testNumpyConversion():
    import numpy

    values = numpy.array([100, 200, 300])
    array = units.Array("length", values=values, unit="m")

    converted_values = array.GetValues("km")
    assert isinstance(converted_values, numpy.ndarray)
    assert list(numpy.array([0.1, 0.2, 0.3])) == list(converted_values)


def testNumpyMultiplicationOperation():
    import numpy

    units_array = units.Array("length", values=[100, 200, 300], unit="m")
    raw_array = [100, 200, 300]
    numpy_array = numpy.array([2.0, 0.5, 0.25])

    units_result = units_array * numpy_array
    raw_result = raw_array * numpy_array

    assert list(raw_result) == [200, 100, 75]
    assert list(units_result.GetValues("m")) == list(raw_result)


def testNoConversion():
    values = [100, 150, 200]
    array = units.Array("length", values=values, unit="m")
    assert array.GetValues() is values


def testNoConversionWhenUsingSameUnit():
    import numpy

    values = numpy.array([100, 150, 200])
    array = units.Array(values, "m")
    assert array.GetValues("m") is values


def testZeroDimensionalNumpyArray():
    import numpy

    values = numpy.array(1000.0)
    array = units.Array(values, "m")
    assert array.GetValues("m") == numpy.array(1000.0)
    assert array.GetValues("km") == numpy.array(1.0)


def testReadOnlyQuantity():
    quantity = ObtainQuantity("m", "length")
    array = units.Array(quantity, values=[1, 2, 3])

    array = array.CreateCopy(unit="km")
    assert array.unit == "km"
    assert approx(array.GetValues()) == [0.001, 0.002, 0.003]


def testCopyPropertiesAndValidation(unit_database_len):
    # Not raises exception because by default validation is False on the copy operation
    array_source = Array("flow rate", values=[-1, -2, -3], unit="m3/s")
    assert not array_source.IsValid()

    array_dest = array_source.CreateCopy()
    assert array_dest.values == [-1, -2, -3]


def testDefaultValues(unit_database_len):
    # Not raises exception because by default validation is False on the copy operation
    array = Array("flow rate")
    assert array.values == []

    array = Array(ObtainQuantity("m"))
    assert array.values == []

    with pytest.raises(AssertionError):
        Array(ObtainQuantity("m"), unit="m")



# -*- coding: cp1252 -*-
from __future__ import absolute_import, unicode_literals

import six
import pytest
from pytest import approx

from barril._foundation.reraise import Reraise
from barril import units
from barril.units import UnitsError
from barril.units.unit_database import UnitDatabase


def testPoscWithoutFillCategories(unit_database_posc_len_no_category):
    unit_database = unit_database_posc_len_no_category

    unit_database.AddCategory("my_len", "length")

    # this category does not exist because we didn't fill the categories
    with pytest.raises(UnitsError):
        units.Scalar("length", 100, "km")

    u1 = units.Scalar("my_len", 100, "km")
    u2 = units.Scalar("my_len", 100000, "m")
    assert approx(abs(u1.value - u2.GetValue("km")), 7) == 0


def testMegagramPerCubicMeterToKilogramPerCubicMeter():
    unit_database = UnitDatabase.CreateDefaultSingleton()
    expected = 1000
    obtained = unit_database.Convert("density", "Mg/m3", "kg/m3", 1)
    assert obtained == expected


def testDivisionError():
    unit_database = UnitDatabase.CreateDefaultSingleton()
    for category, category_info in six.iteritems(
        unit_database.categories_to_quantity_types
    ):
        base_unit = category_info.default_unit
        for unit in unit_database.GetValidUnits(category):
            for i in [-1, 0, 1]:
                try:
                    unit_database.Convert(category, base_unit, unit, i)
                except Exception as e:
                    Reraise(e, "Error converting: from: %s to: %s" % (base_unit, unit))

                try:
                    unit_database.Convert(category, unit, base_unit, i)
                except Exception as e:
                    Reraise(e, "Error converting: from: %s to: %s" % (unit, base_unit))


def testScfPerBblToM3ToM3():
    unit_database = UnitDatabase.CreateDefaultSingleton()
    expected = 0.17776487178535644
    obtained = unit_database.Convert(
        "standard volume per volume", "scf/bbl", "scm(15C)/m3", 1.0
    )
    assert approx(abs(obtained - expected), 7) == 0



from __future__ import absolute_import, division, unicode_literals

import operator

import six
import pytest

from barril.units import ObtainQuantity
from barril.units.unit_database import UnitDatabase


def testOperation():
    unit_database = UnitDatabase.GetSingleton()
    unit_database.CheckDefaultUnitDatabase()
    q = ObtainQuantity(unit="m", category="length")

    SCALAR_OPERATION = [
        operator.add,
        operator.sub,
        operator.truediv,
        operator.mul,
        operator.mod,
    ]
    SCALAR_TYPES = six.integer_types + (float,)

    for operation in SCALAR_OPERATION:
        for cast_type in SCALAR_TYPES:
            # Left
            q2 = operation(cast_type(2), q)
            assert q2.GetQuantityType() == "length"
            assert q2.GetUnit() == "m"

            # Right
            q2 = operation(q, cast_type(2))
            assert q2.GetQuantityType() == "length"
            assert q2.GetUnit() == "m"

    q2 = abs(q)
    assert q2.GetQuantityType() == "length"
    assert q2.GetUnit() == "m"

    q2 = q + q
    assert q2.GetQuantityType() == "length"
    assert q2.GetUnit() == "m"

    q2 = q - q
    assert q2.GetQuantityType() == "length"
    assert q2.GetUnit() == "m"

    q2 = q * int(2)
    assert q2.GetQuantityType() == "length"
    assert q2.GetUnit() == "m"

    q2 = q * 2.0
    assert q2.GetQuantityType() == "length"
    assert q2.GetUnit() == "m"

    q2 = q * q
    assert q2.GetQuantityType() == "(length) ** 2"
    assert q2.GetUnit() == "m2"

    q2 = q * q * q
    assert q2.GetQuantityType() == "(length) ** 3"
    assert q2.GetUnit() == "m3"

    # Check pow
    q2 = q ** 1
    assert q2.GetQuantityType() == "length"
    assert q2.GetUnit() == "m"

    q2 = q ** 2
    assert q2.GetQuantityType() == "(length) ** 2"
    assert q2.GetUnit() == "m2"

    q2 = q ** 3
    assert q2.GetQuantityType() == "(length) ** 3"
    assert q2.GetUnit() == "m3"

    q2 = pow(q, 3)
    assert q2.GetQuantityType() == "(length) ** 3"
    assert q2.GetUnit() == "m3"

    q2 = (q * q) / q
    assert q2.GetQuantityType() == "length"
    assert q2.GetUnit() == "m"

    q2 = (q * q) / ObtainQuantity(unit="d", category="time")
    assert q2.GetQuantityType() == "(length) ** 2 / time"
    assert q2.GetUnit() == "m2/d"

    with pytest.raises(TypeError):
        operator.mul(q, "s")
